from __future__ import annotations

import logging
from contextlib import suppress
from typing import TYPE_CHECKING

from better_translation._babel.extractor import extract_from_dir

if TYPE_CHECKING:
    from better_translation.integrations.django.models import BaseMessage


logger = logging.getLogger(__name__)


def update_messages(
    base_dir: str,
    message_model: type[BaseMessage],
) -> None:
    """Update messages in database.

    :param base_dir: Base directory for search.
    :param message_model: Message model.
    """
    already_saved_messages = message_model.objects.all()
    already_saved_messages_by_raw_singular = {
        message.raw_singular: message for message in already_saved_messages
    }
    already_saved_messages_by_raw_plural = {
        message.raw_plural: message for message in already_saved_messages
    }
    unused_message_ids = {message.id for message in already_saved_messages}

    messages_to_save: list[BaseMessage] = []
    messages_to_update: list[BaseMessage] = []

    extracted_messages = extract_from_dir(base_dir)

    for extracted_message in extracted_messages:
        logger.debug("Extracted message: %s", extracted_message)
        _filename, _lineno, messages, _comments, _context = extracted_message

        if isinstance(messages, str):
            raw_singular, raw_plural = messages, ""
        else:
            raw_singular, raw_plural = messages

        already_saved_message = already_saved_messages_by_raw_singular.get(
            raw_singular,
        ) or already_saved_messages_by_raw_plural.get(raw_plural)

        if already_saved_message is None:
            logger.info("New message: %s", raw_singular)
            message = message_model(
                raw_singular=raw_singular,
                raw_plural=raw_plural,
            )
            messages_to_save.append(message)
        else:
            logger.debug("Existing message: %s", already_saved_message)

            message_to_update = message_model(
                id=already_saved_message.id,
                raw_singular=raw_singular,
                raw_plural=raw_plural,
            )

            with suppress(KeyError):
                unused_message_ids.remove(already_saved_message.id)

            if not already_saved_message.is_used:
                message_to_update.is_used = True

            if (
                message_to_update.raw_singular
                != already_saved_message.raw_singular
                or message_to_update.raw_plural
                != already_saved_message.raw_plural
                or message_to_update.is_used != already_saved_message.is_used
            ):
                logger.info("Message to update: %s", message_to_update)
                messages_to_update.append(message_to_update)

    logger.info("Messages to save: %s", messages_to_save)
    logger.info("Messages to update: %s", messages_to_update)

    message_model.objects.bulk_create(messages_to_save)
    message_model.objects.bulk_update(
        messages_to_update,
        ["raw_singular", "raw_plural", "is_used"],
    )

    message_model.objects.filter(id__in=unused_message_ids).update(
        is_used=False,
    )



from __future__ import annotations

from typing import cast
from uuid import uuid4

try:
    from django.db import models
except ImportError as e:
    msg = "Django is not installed. To use Django integration, install Django."
    raise ValueError(msg) from e


MAX_MESSAGE_LENGTH = 4096


class DefaultLocalesChoices(models.TextChoices):
    RU = "ru", "Russian"
    EN = "en", "English"


class BaseMessage(models.Model):
    id = models.UUIDField(
        primary_key=True,
        default=uuid4,
        editable=False,
    )

    raw_singular = models.CharField(
        max_length=MAX_MESSAGE_LENGTH,
        verbose_name="Raw singular form of the message",
        unique=True,
    )

    raw_plural = models.CharField(
        max_length=MAX_MESSAGE_LENGTH,
        verbose_name="Raw plural form of the message",
        default="",
    )

    is_used = models.BooleanField(
        verbose_name="Is used",
        default=True,
    )

    translations: models.QuerySet[BaseTranslation]
    """A QuerySet of translations for this message."""

    def __str__(self) -> str:
        return f"Message: {self.raw_singular}"

    class _Meta:
        abstract = True

    Meta = cast("type[models.Model.Meta]", _Meta)


class BaseTranslation(models.Model):
    locale = models.CharField(
        max_length=5,
        verbose_name="Locale",
        choices=DefaultLocalesChoices.choices,
    )
    singular = models.TextField(
        verbose_name="Singular form of the text",
    )
    plural = models.TextField(
        verbose_name="Plural form of the text",
        default="",
        blank=True,
    )

    message = models.ForeignKey(
        BaseMessage,
        on_delete=models.CASCADE,
        related_name="translations",
    )

    def __str__(self) -> str:
        return f"MessageTranslation({self.locale})"

    class _Meta:
        abstract = True

    Meta = cast("type[models.Model.Meta]", _Meta)



from __future__ import annotations

from django.contrib import admin

from better_translation.integrations.django.models import (
    BaseMessage,
    BaseTranslation,
)

admin.ModelAdmin.__class_getitem__ = classmethod(  # type: ignore[assignment]
    lambda cls, _: cls,
)
admin.TabularInline.__class_getitem__ = classmethod(  # type: ignore[assignment]
    lambda cls, _: cls,
)


class BaseMessageAdmin(admin.ModelAdmin[BaseMessage]):
    list_display = ("raw_singular", "is_used")
    readonly_fields = ("raw_singular", "raw_plural", "is_used")
    search_fields = (
        "raw_singular",
        "raw_plural",
        "translations__singular",
        "translations__plural",
    )


class BaseTranslationInline(admin.TabularInline[BaseTranslation]):
    extra = 0



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

T = TypeVar("T", bound=Any)


class ICache(ABC, Generic[T]):
    @abstractmethod
    def get(self, *args_keys: Any, **kwargs_keys: Any) -> T | None:
        ...

    @abstractmethod
    def set(self, value: T, *args_keys: Any, **kwargs_keys: Any) -> None:
        ...

    @abstractmethod
    def clear(self) -> None:
        ...


class InMemoryCache(ICache[T]):
    def __init__(self) -> None:
        self._cache: dict[tuple[Any, ...], Any] = {}

    def get(self, *args_keys: Any, **kwargs_keys: Any) -> T | None:
        return self._cache.get((*args_keys, *kwargs_keys.items()))

    def set(self, value: T, *args_keys: Any, **kwargs_keys: Any) -> None:
        self._cache[(*args_keys, *kwargs_keys.items())] = value

    def clear(self) -> None:
        self._cache.clear()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, cast

from better_translation.types import (
    RawPlural,
    RawSingular,
    TranslatedPlural,
    TranslatedSingular,
    TranslatedText,
)

if TYPE_CHECKING:
    from better_translation.types import Locale


class ITranslator(ABC):
    @abstractmethod
    def translate(
        self,
        locale: Locale,
        raw_singular: RawSingular,
        raw_plural: RawPlural | None = None,
        n: int = 1,
    ) -> TranslatedText:
        ...


class DefaultTranslator(ITranslator):
    def translate(
        self,
        locale: Locale,
        raw_singular: RawSingular,
        raw_plural: RawPlural | None = None,
        n: int = 1,
    ) -> TranslatedText:
        if raw_plural is None:
            return self._translate_singular(locale, raw_singular)

        return self._translate_plural(
            locale,
            raw_singular,
            raw_plural,
            n,
        )

    def _translate_singular(
        self,
        locale: Locale,  # noqa: ARG002
        raw_singular: RawSingular,
    ) -> TranslatedSingular:
        return cast(TranslatedSingular, raw_singular)

    def _translate_plural(
        self,
        locale: Locale,  # noqa: ARG002
        raw_singular: RawSingular,
        raw_plural: RawPlural,
        n: int,
    ) -> TranslatedPlural:
        return cast(
            TranslatedPlural,
            raw_singular if n == 1 else raw_plural,
        )



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from better_translation.cache import ICache, InMemoryCache
from better_translation.integrations.django.models import (
    BaseMessage,
    BaseTranslation,
)
from better_translation.translator import ITranslator

if TYPE_CHECKING:
    from better_translation.types import (
        Locale,
        RawPlural,
        RawSingular,
        TranslatedText,
    )

logger = logging.getLogger(__name__)


@dataclass(slots=True, frozen=True)
class DjangoTranslator(ITranslator):
    message_model: type[BaseMessage]
    translations_cache: ICache[BaseTranslation | None] = field(
        default_factory=InMemoryCache[BaseTranslation | None],
    )

    def translate(
        self,
        locale: Locale,
        raw_singular: RawSingular,
        raw_plural: RawPlural | None = None,
        n: int = 1,
    ) -> TranslatedText:
        translation = self._get_translation(raw_singular, locale)

        if raw_plural:
            return self._translate_plural(
                raw_singular,
                raw_plural,
                n,
                translation,
            )

        return self._translate_singular(
            raw_singular,
            translation,
        )  # type: ignore[return-value]

    def _translate_singular(
        self,
        raw_singular: RawSingular,
        translation: BaseTranslation | None,
    ) -> str:
        return translation.singular if translation else raw_singular

    def _translate_plural(
        self,
        raw_singular: RawSingular,
        raw_plural: RawPlural | None,
        n: int,
        translation: BaseTranslation | None,
    ) -> TranslatedText:
        if translation:
            singular = translation.singular
            plural = translation.plural
        else:
            singular = raw_singular
            plural = raw_plural

        if plural:
            return plural if n != 1 else singular  # type: ignore[return-value]

        return singular  # type: ignore[return-value]

    def _get_message(self, raw_singular: RawSingular) -> BaseMessage:
        logger.debug("Getting message %s", raw_singular)
        message, _ = self.message_model.objects.get_or_create(
            raw_singular=raw_singular,
        )

        return message

    def _get_translation(
        self,
        raw_singular: RawSingular,
        locale: Locale,
    ) -> BaseTranslation | None:
        cached_translation = self.translations_cache.get(
            raw_singular,
            locale,
        )

        if cached_translation:
            logger.debug(
                "Returning translation %s from cache",
                cached_translation,
            )
            return cached_translation

        logger.debug("Getting translation for %s", raw_singular)

        message = self._get_message(raw_singular)
        translation = message.translations.filter(
            locale=locale,
        ).first()

        self.translations_cache.set(
            translation,
            raw_singular,
            locale,
        )

        return translation



from __future__ import annotations

import codecs
import os
import re
from typing import IO

# Regexp to match python magic encoding line
PYTHON_MAGIC_COMMENT_re = re.compile(
    rb"[ \t\f]* \# .* coding[=:][ \t]*([-\w.]+)",
    re.VERBOSE,
)


def parse_encoding(fp: IO[bytes]) -> str | None:
    """Deduce the encoding of a source file from magic comment.

    It does this in the same way as the `Python interpreter`__

    .. __: https://docs.python.org/3.4/reference/lexical_analysis.html#encoding-declarations

    The ``fp`` argument should be a seekable file object.

    (From Jeff Dairiki)
    """
    pos = fp.tell()
    fp.seek(0)
    try:
        line1 = fp.readline()
        has_bom = line1.startswith(codecs.BOM_UTF8)
        if has_bom:
            line1 = line1[len(codecs.BOM_UTF8) :]

        m = PYTHON_MAGIC_COMMENT_re.match(line1)
        if not m:
            try:
                import ast

                ast.parse(line1.decode("latin-1"))
            except (ImportError, SyntaxError, UnicodeEncodeError):
                # Either it's a real syntax error, in which case the source is
                # not valid python source, or line2 is a continuation of line1,
                # in which case we don't want to scan line2 for a magic
                # comment.
                pass
            else:
                line2 = fp.readline()
                m = PYTHON_MAGIC_COMMENT_re.match(line2)

        if has_bom:
            if m:
                magic_comment_encoding = m.group(1).decode("latin-1")
                if magic_comment_encoding != "utf-8":
                    raise SyntaxError(
                        "encoding problem: {} with BOM".format(
                            magic_comment_encoding,
                        ),
                    )
            return "utf-8"
        elif m:
            return m.group(1).decode("latin-1")
        else:
            return None
    finally:
        fp.seek(pos)


PYTHON_FUTURE_IMPORT_re = re.compile(r"from\s+__future__\s+import\s+\(*(.+)\)*")


def parse_future_flags(fp: IO[bytes], encoding: str = "latin-1") -> int:
    """Parse the compiler flags by :mod:`__future__` from the given Python
    code.
    """
    import __future__

    pos = fp.tell()
    fp.seek(0)
    flags = 0
    try:
        body = fp.read().decode(encoding)

        # Fix up the source to be (hopefully) parsable by regexpen.
        # This will likely do untoward things if the source code itself is broken.

        # (1) Fix `import (\n...` to be `import (...`.
        body = re.sub(r"import\s*\([\r\n]+", "import (", body)
        # (2) Join line-ending commas with the next line.
        body = re.sub(r",\s*[\r\n]+", ", ", body)
        # (3) Remove backslash line continuations.
        body = re.sub(r"\\\s*[\r\n]+", " ", body)

        for m in PYTHON_FUTURE_IMPORT_re.finditer(body):
            names = [x.strip().strip("()") for x in m.group(1).split(",")]
            for name in names:
                feature = getattr(__future__, name, None)
                if feature:
                    flags |= feature.compiler_flag
    finally:
        fp.seek(pos)
    return flags


def pathmatch(pattern: str, filename: str) -> bool:
    """Extended pathname pattern matching.

    This function is similar to what is provided by the ``fnmatch`` module in
    the Python standard library, but:

     * can match complete (relative or absolute) path names, and not just file
       names, and
     * also supports a convenience pattern ("**") to match files at any
       directory level.

    Examples
    --------
    >>> pathmatch('**.py', 'bar.py')
    True
    >>> pathmatch('**.py', 'foo/bar/baz.py')
    True
    >>> pathmatch('**.py', 'templates/index.html')
    False

    >>> pathmatch('./foo/**.py', 'foo/bar/baz.py')
    True
    >>> pathmatch('./foo/**.py', 'bar/baz.py')
    False

    >>> pathmatch('^foo/**.py', 'foo/bar/baz.py')
    True
    >>> pathmatch('^foo/**.py', 'bar/baz.py')
    False

    >>> pathmatch('**/templates/*.html', 'templates/index.html')
    True
    >>> pathmatch('**/templates/*.html', 'templates/foo/bar.html')
    False

    :param pattern: the glob pattern
    :param filename: the path name of the file to match against
    """
    symbols = {
        "?": "[^/]",
        "?/": "[^/]/",
        "*": "[^/]+",
        "*/": "[^/]+/",
        "**/": "(?:.+/)*?",
        "**": "(?:.+/)*?[^/]+",
    }

    if pattern.startswith("^"):
        buf = ["^"]
        pattern = pattern[1:]
    elif pattern.startswith("./"):
        buf = ["^"]
        pattern = pattern[2:]
    else:
        buf = []

    for idx, part in enumerate(re.split("([?*]+/?)", pattern)):
        if idx % 2:
            buf.append(symbols[part])
        elif part:
            buf.append(re.escape(part))
    match = re.match("".join(buf) + "$", filename.replace(os.sep, "/"))
    return match is not None



from __future__ import annotations

from typing import NewType

Locale = NewType("Locale", str)

RawText = NewType("RawText", str)
RawSingular = NewType("RawSingular", RawText)
RawPlural = NewType("RawPlural", RawText)

TranslatedText = NewType("TranslatedText", str)
TranslatedSingular = NewType("TranslatedSingular", TranslatedText)
TranslatedPlural = NewType("TranslatedPlural", TranslatedText)
